import { FormGroup, FormArray, Validators, FormBuilder } from '@angular/forms';
import { Injectable } from '@angular/core';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 * @param {?} factory
 * @param {?=} destroy
 * @return {?}
 */
function lazy(factory, destroy) {
    /** @type {?} */
    let init = false;
    /** @type {?} */
    let lVal;
    /** @type {?} */
    let err;
    /**
     * @return {?}
     */
    function value() {
        if (!init) {
            init = true;
            try {
                lVal = factory();
            }
            catch (e) {
                err = e;
            }
        }
        if (err) {
            throw err;
        }
        return lVal;
    }
    value.reset = (/**
     * @return {?}
     */
    () => {
        if (init) {
            err = undefined;
            if (typeof destroy === 'function') {
                destroy(lVal);
            }
            init = false;
        }
    });
    value.isInit = (/**
     * @return {?}
     */
    () => init);
    return value;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} control
 * @param {...?} ps
 * @return {?}
 */
function traverseControl(control, ...ps) {
    if (ps.length === 0 || !control) {
        return control;
    }
    if (control instanceof FormGroup || control instanceof FormArray) {
        const [p, ...lps] = ps;
        return traverseControl(control.controls[p], ...lps);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} ctrls
 * @return {?}
 */
function getElements(ctrls) {
    return Array.isArray(ctrls) ? ctrls :
        typeof ctrls === 'object' && Object.values(ctrls) || [];
}
/**
 * @param {?} root
 * @param {?} action
 * @param {?} updateState
 * @return {?}
 */
function tapControls(root, action, updateState) {
    for (const ctrl of getElements(root && root.controls)) {
        if (ctrl instanceof FormGroup || ctrl instanceof FormArray) {
            tapControls(ctrl, action, updateState);
            continue;
        }
        if (typeof action === 'function') {
            action(ctrl);
        }
        if (updateState) {
            ctrl.updateValueAndValidity();
        }
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 * @param {?} items
 * @return {?}
 */
function uniq(items) {
    return (items || []).filter((/**
     * @param {?} val
     * @param {?} idx
     * @param {?} self
     * @return {?}
     */
    (val, idx, self) => val && self.indexOf(val) === idx));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class DependentValidatorBuilderImpl {
    /**
     * @param {?} affectedControls
     * @param {?} validate
     */
    constructor(affectedControls, validate) {
        this.affectedControls = affectedControls;
        this.validate = validate;
    }
    /**
     * @return {?}
     */
    inverse() {
        return new DependentValidatorBuilderImpl(this.affectedControls, (/**
         * @return {?}
         */
        () => !this.validate()));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class DependentValidatorBuilderRootImpl {
    /**
     * @param {?} form
     */
    constructor(form) {
        this.form = form;
    }
    /**
     * @param {...?} builders
     * @return {?}
     */
    all(...builders) {
        return new DependentValidatorBuilderImpl((/**
         * @return {?}
         */
        () => uniq([].concat(builders.map((/**
         * @param {?} b
         * @return {?}
         */
        b => b.affectedControls()))))), (/**
         * @return {?}
         */
        () => builders.every((/**
         * @param {?} b
         * @return {?}
         */
        b => b.validate()))));
    }
    /**
     * @param {...?} builders
     * @return {?}
     */
    any(...builders) {
        return new DependentValidatorBuilderImpl((/**
         * @return {?}
         */
        () => uniq([].concat(builders.map((/**
         * @param {?} b
         * @return {?}
         */
        b => b.affectedControls()))))), (/**
         * @return {?}
         */
        () => builders.some((/**
         * @param {?} b
         * @return {?}
         */
        b => b.validate()))));
    }
    /**
     * @param {?} predicate
     * @param {?} firstPath
     * @param {...?} paths
     * @return {?}
     */
    validator(predicate, firstPath, ...paths) {
        /** @type {?} */
        const formFn = lazy((/**
         * @return {?}
         */
        () => this.form()));
        /** @type {?} */
        const ctrlFn = lazy((/**
         * @return {?}
         */
        () => {
            if (typeof firstPath === 'string' || typeof firstPath === 'number') {
                return traverseControl(formFn(), firstPath, ...paths);
            }
            else if (typeof firstPath === 'function') {
                return firstPath(formFn());
            }
        }));
        return new DependentValidatorBuilderImpl((/**
         * @return {?}
         */
        () => [ctrlFn()]), (/**
         * @return {?}
         */
        () => {
            /** @type {?} */
            const ctrl = ctrlFn();
            return ctrl && typeof predicate === 'function' && predicate(ctrl.value, formFn());
        }));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class DependentValidatorImpl {
    /**
     * @param {?} rootFn
     * @param {?} builderFn
     */
    constructor(rootFn, builderFn) {
        this.rootFn = rootFn;
        this.builderFn = builderFn;
        this.builder = lazy((/**
         * @return {?}
         */
        () => {
            if (typeof this.builderFn === 'function') {
                /** @type {?} */
                const builderRoot = new DependentValidatorBuilderRootImpl((/**
                 * @return {?}
                 */
                () => this.rootFn()));
                return this.builderFn(builderRoot);
            }
        }));
        this.subscribedControls = [];
        this.subscription = lazy((/**
         * @return {?}
         */
        () => this.builder().affectedControls()
            .filter((/**
         * @param {?} item
         * @return {?}
         */
        item => !!item)).map((/**
         * @param {?} item
         * @return {?}
         */
        item => item
            .valueChanges.subscribe((/**
         * @return {?}
         */
        () => this.subscribedControls
            .forEach((/**
         * @param {?} control
         * @return {?}
         */
        control => control.updateValueAndValidity()))))))), (/**
         * @param {?} subs
         * @return {?}
         */
        subs => subs.forEach((/**
         * @param {?} s
         * @return {?}
         */
        s => s.unsubscribe()))));
    }
    /**
     * @return {?}
     */
    reset() {
        this.builder.reset();
        this.subscription.reset();
    }
    /**
     * @template THIS
     * @this {THIS}
     * @param {?} path
     * @param {...?} validators
     * @return {THIS}
     */
    applyValidator(path, ...validators) {
        /** @type {?} */
        const paths = Array.isArray(path) ? path : [path];
        /** @type {?} */
        const ctrl = traverseControl((/** @type {?} */ (this)).rootFn(), ...paths);
        if (ctrl) {
            ctrl.setValidators((/** @type {?} */ (this)).createValidator(...validators));
        }
        return (/** @type {?} */ (this));
    }
    /**
     * @param {...?} validators
     * @return {?}
     */
    createValidator(...validators) {
        /** @type {?} */
        const validate = Validators.compose(validators);
        /** @type {?} */
        let ctrlSubscribed = false;
        return (/**
         * @param {?} ctrl
         * @return {?}
         */
        ctrl => {
            /** @type {?} */
            let builder;
            if (!ctrlSubscribed) {
                ctrlSubscribed = true;
                if (this.subscribedControls.indexOf(ctrl) === -1) {
                    this.subscribedControls.push(ctrl);
                }
            }
            if (!this.subscription.isInit() && this.rootFn()) {
                builder = this.builder();
                this.subscription();
            }
            else if (this.builder.isInit()) {
                builder = this.builder();
            }
            return (!builder || builder.validate()) ? validate(ctrl) : null;
        });
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class FormUtils {
    /**
     * @param {?} rootFn
     */
    constructor(rootFn) {
        this.rootFn = rootFn;
        this._root = lazy((/**
         * @return {?}
         */
        () => {
            if (this.rootFn instanceof FormGroup || this.rootFn instanceof FormArray) {
                return this.rootFn;
            }
            else if (typeof this.rootFn === 'function') {
                return this.rootFn();
            }
        }));
    }
    /**
     * @return {?}
     */
    get root() {
        /** @type {?} */
        const r = this._root;
        if (!r()) {
            r.reset();
        }
        return (/** @type {?} */ (r()));
    }
    /**
     * @param {?=} action
     * @param {?=} updateState
     * @return {?}
     */
    tapControls(action, updateState = true) {
        tapControls(this.root, action, updateState);
    }
    /**
     * @param {?} builderFn
     * @return {?}
     */
    createDependentValidator(builderFn) {
        return new DependentValidatorImpl((/**
         * @return {?}
         */
        () => this.root), builderFn);
    }
    /**
     * @return {?}
     */
    makePristine() {
        this.tapControls((/**
         * @param {?} ctrl
         * @return {?}
         */
        ctrl => {
            ctrl.markAsPristine();
            ctrl.markAsUntouched();
        }));
    }
    /**
     * @return {?}
     */
    makeDirty() {
        this.tapControls((/**
         * @param {?} ctrl
         * @return {?}
         */
        ctrl => {
            ctrl.markAsDirty();
            ctrl.markAsTouched();
        }));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template T
 */
class ExtendedFormBuilderResult {
    /**
     * @param {?} fb
     * @param {?} factory
     */
    constructor(fb, factory) {
        this.fb = fb;
        this.factory = factory;
        this.validators = [];
        this._formUtils = lazy((/**
         * @return {?}
         */
        () => new FormUtils((/**
         * @return {?}
         */
        () => this.form))));
        this.form = this.factory(this.fb, (/**
         * @param {?} dv
         * @return {?}
         */
        dv => {
            /** @type {?} */
            const v = this.formUtils.createDependentValidator(dv);
            this.validators.push(v);
            return v;
        }));
    }
    /**
     * @return {?}
     */
    get formUtils() {
        return this._formUtils();
    }
    /**
     * @return {?}
     */
    updateValidators() {
        this.validators.forEach((/**
         * @param {?} item
         * @return {?}
         */
        item => item.reset()));
        this.formUtils.tapControls();
    }
}
class ExtendedFormBuilder {
    /**
     * @param {?} fb
     */
    constructor(fb) {
        this.fb = fb;
    }
    /**
     * @template T
     * @param {?} factory
     * @return {?}
     */
    build(factory) {
        return new ExtendedFormBuilderResult(this.fb, factory);
    }
}
ExtendedFormBuilder.decorators = [
    { type: Injectable }
];
/** @nocollapse */
ExtendedFormBuilder.ctorParameters = () => [
    { type: FormBuilder }
];

export { ExtendedFormBuilder, FormUtils, tapControls, traverseControl };
//# sourceMappingURL=ngx-reactive-form.js.map
